/* Copyright (C) 2009-2010 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * If the integration of miniscope code requires the use of a license
 * other than GPL, an alternative license may be arranged.
 */

#include "BusPirateScope.h"
#include "..\miniscope\device\MiniscopeDevice.h"
#include <windows.h>
#include <stdio.h>
#include <assert.h>
#include <algorithm>

#ifdef TEST
#   include "TestSignal.h"
#endif

extern int iComPortNumber;

DWORD WINAPI ThreadProc(LPVOID threaddata)
{
    enum { SAMPLES_COUNT = 128 };
    short samples[SAMPLES_COUNT];
    unsigned char data[SAMPLES_COUNT];
    busPirateScope->log("RX thread created");
    while (busPirateScope->connected)
    {
#if 1
        // get samples from device
        DWORD readed;
        bool status = ReadFile(busPirateScope->hCommDev, data, sizeof(data), &readed, NULL);
        if (status == false)
        {
            busPirateScope->connected = false;
            if (busPirateScope->connected && busPirateScope->callbackConnect)
                busPirateScope->callbackConnect(DEVICE_DISCONNECTED, (char*)"ReadFile failed, disconnected");
            busPirateScope->log("ReadFile error, RX thread exit");

            busPirateScope->ExitContinuousSampling();

            return 0;
        }
        //busPirateScope->log("RX %d", readed);

        int total_samples = readed;

        if (total_samples == 0)
            continue;

        for (int i=0; i<total_samples; i++)
        {
            samples[i] = data[i];
        }

#else
        int total_samples = rand()%SAMPLES_COUNT;
        memcpy(samples, testSignal.Generate(total_samples, 1.0/5720.0, 6.6/1024.0), total_samples*sizeof(samples[0]));
        Sleep(50);
#endif

        busPirateScope->OnDataReceive(samples, total_samples);
    }
    busPirateScope->ExitContinuousSampling();
    busPirateScope->log("RX thread exit");
    return 0;
}

BusPirateScope::BusPirateScope():
    bRunning(false),
    mode(TRIGGER_MODE_AUTO),
    bManualTriggerRequest(false),
    lastTriggerDistance(0),
    frame_position(0),
    Thread(NULL),
    hCommDev(INVALID_HANDLE_VALUE)
{
    // some default values
    settings.iChannelCount = 0; // this will allow to ignore setting array pointers inside settings structure to valid values
    settings.iSamplingPeriod = 1;
    settings.eTriggerMode = TRIGGER_MODE_AUTO;
    settings.eTriggerType = TRIGGER_SLOPE_UP;
    settings.iTriggerLevel = 0;
    settings.iBufferSize = 4096;

    upcoming_settings = settings;
    callbackData = NULL;
    callbackLog = NULL;
    callbackConnect = NULL;

    OutputDebugStr("busPirateScope created");
}

BusPirateScope::~BusPirateScope()
{
    Disconnect();
}

void BusPirateScope::log(const char* const lpData, ...)
{
    static char buf[256]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int BusPirateScope::Connect(void)
{
#ifdef TEST
    ConnectCb(DEVICE_CONNECTED, (char*)"Bus Pirate (test)");
    log("Test run!");
    connected = true;
    DWORD dwtid;
    if (!Thread)
        Thread = CreateThread(NULL, 0, ThreadProc, this, 0, &dwtid);
    SetThreadPriority(Thread, THREAD_PRIORITY_TIME_CRITICAL);
    return 0;
#endif

    if (hCommDev != INVALID_HANDLE_VALUE)
        CloseHandle(hCommDev);
    hCommDev = INVALID_HANDLE_VALUE;

    COMMTIMEOUTS ctmoNew, ctmoOld;
    char szComPort[32];
    sprintf(szComPort, "\\\\.\\COM%d", iComPortNumber);
    hCommDev = CreateFile(szComPort, GENERIC_READ | GENERIC_WRITE, 0,
                          NULL, OPEN_EXISTING, 0, 0);
    if (hCommDev != INVALID_HANDLE_VALUE)
    {
        //ustawienia portu
        if (SetupComm(hCommDev, 8192, 4096) == false)
        {
            log("SetupComm call failed!");
        }
        //ustawienia timeoutow
        GetCommTimeouts(hCommDev,&ctmoOld);
        //maximum time, in milliseconds, allowed to elapse between the arrival of two characters
        ctmoNew.ReadIntervalTimeout = 100;
        ctmoNew.ReadTotalTimeoutConstant = 40; //czas dodawany dla calej operacji
        ctmoNew.ReadTotalTimeoutMultiplier = 10;   //to jest mnozone przez liczbe bajtow dla operacji IO
        ctmoNew.WriteTotalTimeoutMultiplier = 0;
        ctmoNew.WriteTotalTimeoutConstant = 600;
        SetCommTimeouts(hCommDev, &ctmoNew);

        // -- ustawienia flag sterujacych DCB --
        dcb.DCBlength = sizeof(dcb);
        GetCommState(hCommDev, &dcb);
        dcb.BaudRate = CBR_115200;
        //jaka i czy jest parzystosc
        dcb.Parity = NOPARITY;
        //liczba bitow stopu
        dcb.StopBits = ONESTOPBIT;
        //wybor liczby bitow danych
        dcb.ByteSize = 8;
        SetCommState(hCommDev, &dcb);
        log( "Opened port %s.", szComPort );
        FlushFileBuffers(hCommDev);

        char text[64];
        sprintf(text, "Reading from COM%d", iComPortNumber);
        ConnectCb(DEVICE_CONNECTED, text);

        connected = true;
        DWORD dwtid;
        if (!Thread)
            Thread = CreateThread(NULL, 0, ThreadProc, this, 0, &dwtid);
        SetThreadPriority(Thread, THREAD_PRIORITY_TIME_CRITICAL);
    }
    else
    {
        ConnectCb(DEVICE_DISCONNECTED, (char*)"Could not connect to port. Check if port number in .cfg file is correct and restart.");
        log("Failed opening %s. Change configuration (.cfg file) and restart.", szComPort);
        return 1;
    }

    return 0;
}

void BusPirateScope::ExitContinuousSampling(void)
{

}

int BusPirateScope::Disconnect(void)
{
    // wait for read thread to terminate
    connected = false;
    Sleep(500);

    if (hCommDev != INVALID_HANDLE_VALUE)
    {
        CloseHandle(hCommDev);
    }
    hCommDev = INVALID_HANDLE_VALUE;

    Thread = NULL;
    return 0;
}

int BusPirateScope::ManualTrigger(void)
{
    bManualTriggerRequest = true;
    return 0;
}

int BusPirateScope::SetTriggerMode(enum E_TRIGGER_MODE m)
{
    mode = m;
    if (mode == TRIGGER_MODE_AUTO)
        lastTriggerDistance = 0;	// reset
    return 0;
}

int BusPirateScope::SetTriggerType(enum E_TRIGGER_TYPE type)
{
    if (type == TRIGGER_SLOPE_UP)
    {
        upcoming_settings.eTriggerType = type;
    }
    else if (type == TRIGGER_SLOPE_DOWN)
    {
        upcoming_settings.eTriggerType = type;
    }
    else
        return -2;
    log("Trigger type set to %d", type);
    return 0;
}

int BusPirateScope::SetTriggerLevel(int level)
{
    //if (level < 0 || level > 1024)
    //    return -1;
    log("Trigger level set to %d", level);
    upcoming_settings.iTriggerLevel = level;
    return 0;
}

int BusPirateScope::Run(bool state)
{
    bRunning = state;
    if (!bRunning)
    {
        bManualTriggerRequest = false;
        lastTriggerDistance = 0;
        frame_position = 0;
    }
    return 0;
}

int BusPirateScope::TestTrigger(short* samples, int len)
{
    int pos = -1;
    int level = settings.iTriggerLevel;
    settings = upcoming_settings;

    if (bManualTriggerRequest || mode == TRIGGER_MODE_CONTINUOUS)
    {
        pos = 0;
    }
    else
    {
        switch (settings.eTriggerType)
        {
        case TRIGGER_SLOPE_UP:
            /** \note oversimplified code; will not detect edge spread across frames */
            for (int i=0; i<len-4; i++)
            {
                if (samples[i] < level && samples[i+1] <= level && samples[i+2] >= level &&
                        samples[i+3] > level)
                {
                    pos = i;
                    break;
                }
            }
            break;
        case TRIGGER_SLOPE_DOWN:
            /** \note oversimplified code; will not detect edge spread across frames  */
            for (int i=0; i<len-4; i++)
            {
                if (samples[i] > level && samples[i+1] >= level && samples[i+2] <= level &&
                        samples[i+3] < level)
                {
                    pos = i;
                    break;
                }
            }
            break;
        default:
            break;
        }
    }
    if (mode == TRIGGER_MODE_AUTO)
    {
        if (pos == -1)
        {
            lastTriggerDistance += len;
            // generate auto-trigger when there is no trigger for the equivalent time of 3 frames
            if (lastTriggerDistance >= 3 * upcoming_settings.iBufferSize)
            {
                pos = 0;
                lastTriggerDistance = 0;
            }
        }
        else
        {
            lastTriggerDistance = 0;
        }
    }
    return pos;
}

void BusPirateScope::OnDataReceive(short samples[], unsigned int total_samples)
{
    short* samples_p = samples;
    int iSamplesCnt = total_samples;

    if (frame_position == 0)
    {
        settings = upcoming_settings;
        //decimator.SetFactor(settings.iSamplingPeriod);
    }

    if (bRunning == false && bManualTriggerRequest == false)
    {
        frame_position = 0;
        return;
    }

    if (frame_position == 0)
    {
        int trigger_pos = TestTrigger(samples, total_samples);
        if (trigger_pos < 0)
        {
            // ignore this data portion
            return;
        }
        else
        {
            samples_p += trigger_pos;
            total_samples -= trigger_pos;
            callbackTrigger();
        }
    }

    if (total_samples + frame_position < settings.iBufferSize)
    {
        callbackData(samples_p, total_samples * sizeof(samples_p[0]));
        frame_position += total_samples;
    }
    else
    {
        if (settings.iBufferSize-frame_position > 0)
            callbackData(samples_p, (settings.iBufferSize-frame_position) * sizeof(samples[0]));
        frame_position += (settings.iBufferSize-frame_position);
        iSamplesCnt -= (settings.iBufferSize-frame_position);
        memmove(samples, samples+(settings.iBufferSize-frame_position),
                iSamplesCnt * sizeof(samples[0]));
        frame_position = iSamplesCnt;
        bManualTriggerRequest = false;
        if (mode != TRIGGER_MODE_CONTINUOUS)
        {
            frame_position = 0;
            settings = upcoming_settings;
        }
        else
        {
            settings = upcoming_settings;
            //log("callbackTrigger, %d samples left", iSamplesCnt);
            callbackTrigger();
            if (iSamplesCnt > 0)
            {
                callbackData(samples, iSamplesCnt * sizeof(samples[0]));
            }
            if (frame_position == 0)
                bManualTriggerRequest = true;
        }

        if (mode == TRIGGER_MODE_SINGLE)
            bRunning = false;
    }
}
